#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <tchar.h>
#include <lsapi/lsapi.h>
#include "hash_table.h"
#include "hash_func.h"

/*

  Perform hopefully MBCS-safe string parsing and conversion.

  This stuff is not yet used as exclusively as it should be.

*/

/***********************************************************
* BOOL Parse_GetHashedToken()                              *
*                         *  *  *  *                       *
*    Arguments:                                            *
*                                                          *
*    - const char *Source;                                 *
*      The input string.                                   *
*    - const char **pNext;                                 *
*      Pointer to be set to after the parsed token.        *
*    - char *OrigCase;                                     *
*      The parsed token as it appears in the source.       *
*    - unsigned *pcchOrigCase;                             *
*      Length of OrigCase on return.                       *
*    - HASH *hash;                                         *
*      Hash of parsed token on return.                     *
*    - char *KeyCase;                                      *
*      The parsed token converted to KeyCase.              *
*    - unsigned *pcchKeyCase;                              *
*      The size of the KeyCase buffer and length of the    *
*      KeyCase token on return.                            *
*                         *  *  *  *                       *
* Parses a token from the input string and returns a hash  *
* value, the token in its original form, and the token     *
* converted to lower case with some other folding done.    *
***********************************************************/
BOOL Parse_GetHashedToken(
  const char *Source, const char **pNext,
  char *OrigCase, unsigned *pcchOrigCase,
  HASH *hash,
  char *KeyCase, unsigned *pcchKeyCase)
{
  unsigned cchKeyCase, cchOrigCase;
  const char *p;

  if (GetToken(Source, OrigCase, pNext, FALSE))
  {
    /* if they specify its size, use that.  otherwise, use MAX_LINE_LENGTH */ 
    /* GetToken requires that OrigCase have room for MAX_LINE_LENGTH, nothing we can do here */ 
    cchKeyCase = (pcchKeyCase && *pcchKeyCase) ? *pcchKeyCase : MAX_LINE_LENGTH;

    cchOrigCase = strlen(OrigCase);
    /* fold to normalized form */ 
    cchKeyCase = FoldString(MAP_FOLDDIGITS|MAP_FOLDCZONE|MAP_PRECOMPOSED,
      OrigCase, cchOrigCase + 1, KeyCase, cchKeyCase);

    if (cchKeyCase)
    {
      --cchKeyCase; /* don't count nul */ 

      /* convert to lowercase to make the hash case insensitive */ 
      CharLowerBuff(KeyCase, cchKeyCase);

      /* hash by TCHAR, not by code point */ 
      HASH_BEGIN(*hash);
      for (p=KeyCase; *p; ++p)
        HASH_ADD(*hash, *p);
      HASH_FINISH(*hash);

      if (pcchOrigCase)
        *pcchOrigCase = cchOrigCase;
      if (pcchKeyCase)
        *pcchKeyCase = cchKeyCase;

      return TRUE;
    }
  }

  if (pcchOrigCase)
    *pcchOrigCase = 0;
  if (pcchKeyCase)
    *pcchKeyCase = 0;

  return FALSE;
}

/***********************************************************
* const char *Parse_SkipSpace()                            *
*                         *  *  *  *                       *
*    Arguments:                                            *
*                                                          *
*    - const char *s;                                      *
*                         *  *  *  *                       *
* Returns a pointer to the first non-whitespace char in s. *
***********************************************************/
const char *Parse_SkipSpace(const char *s)
{
  /* _ismbcspace is the same as isspace */ 

  while (isspace((unsigned char)*s)) ++s;
  return s;
}

/* get static string length at compile time */ 
#define CT_STRLEN(STRING) (sizeof(STRING)/sizeof(*STRING)-1)

/***********************************************************
* unsigned Parse_ReplaceSearchArgs()                       *
*                         *  *  *  *                       *
*    Arguments:                                            *
*                                                          *
*    - const char *URL;                                    *
*      The URL template.                                   *
*    - const char *Args;                                   *
*      The argument string.                                *
*    - char *Buffer;                                       *
*      Where to store the result.                          *
*    - unsigned cchCapacity;                               *
*      Size of Buffer in chars.                            *
*                         *  *  *  *                       *
* Copies the URL template to Buffer, with all occurances   *
* of the string "<args>" replaced with an escaped version  *
* of Args.                                                 *
***********************************************************/
unsigned Parse_ReplaceSearchArgs(const char *URL, const char *Args, char *Buffer, unsigned cchCapacity)
{
  unsigned cchUsed = 0;
  const char *p;

  while (*URL)
  {
    /* look for replacement string */ 
    if (*URL == '<' && CSTR_EQUAL ==
      CompareString(LOCALE_USER_DEFAULT, NORM_IGNORECASE,
        URL, CT_STRLEN("<args>"), "<args>", CT_STRLEN("<args>")))
    {
      /* replace */ 
      URL += CT_STRLEN("<args>");

      p = Args;
      while (*p)
      {
/*
        can't convert to UTF-8 because:
          Win9x doesn't support conversion to/from UCS-2 without Microsoft Layer for Unicode
          Win95/NT3.x don't support conversion to UTF-8 at all
          there are lots of input encodings so I don't want to do it manually
          don't know that the site wants UTF-8 anyway (Google seems to like it, but it has ie= to set the input encoding)

        so I guess we just leave it in the current encoding and hope for the best.
*/
        /* non-ASCII alphanumerics don't count */ 
        if (!isalnum((unsigned char)*p))
        {
          /* there are a few (~9) non-alhpanumeric chars that can occur unescaped */ 
          /* but according to RFC 2396 it shouldn't hurt to escape them           */ 
          *Buffer++ = '%';
          if (++cchUsed == cchCapacity)
            break;
          *Buffer++ = "0123456789ABCDEF"[(unsigned char)(*p) >> 4];
          if (++cchUsed == cchCapacity)
            break;
          *Buffer++ = "0123456789ABCDEF"[(unsigned char)(*p) & 0x0F];
          if (++cchUsed == cchCapacity)
            break;
          ++p;

          /* typically spaces are converted to +, but that's service-specific */ 
          /* and it's easier to just do normal escape for everything.         */ 
        }
        else
        {
          /* lead/trail should be encoded separately. */ 
          *Buffer++ = *p++;
          if (++cchUsed == cchCapacity)
            break;
        }
      }
      if (cchUsed > cchCapacity)
        break;
    }
    else
    {
      if (IsDBCSLeadByte(*URL))
      {
        /* copy an extra char, ignoring '<' as a trail byte */ 
        *Buffer++ = *URL++;
        if (++cchUsed == cchCapacity)
          break;
      }
      *Buffer++ = *URL++;
      if (++cchUsed == cchCapacity)
        break;
    }
  }
  *Buffer = 0;
  if (cchUsed > cchCapacity)
    return 0;
  return cchUsed;
}

/***********************************************************
* char *Parse_FindLastFileToken()                          *
*                         *  *  *  *                       *
*    Arguments:                                            *
*                                                          *
*    - char *sBuffer;                                      *
*      The input string.                                   *
*    - unsigned cchPrefix;                                 *
*      Maximum number of chars to read in sBuffer.         *
*                         *  *  *  *                       *
* Find the beginning of the last "file token," which may   *
* be delimited by spaces or ""s, but not by single quotes  *
* or brackets.                                             *
***********************************************************/
char *Parse_FindLastFileToken(char *sBuffer, unsigned cchPrefix)
{
  BOOL bInToken = FALSE;
  char cQuote = 0;
  char *sLastTokenStart = NULL;
  while (--cchPrefix)
  {
    if (cQuote)
    {
      if (*sBuffer == cQuote)
      {
        cQuote = 0;
        bInToken = FALSE;
      }
    }
    else if (*sBuffer == '"')
    {
      cQuote = *sBuffer;
      bInToken = TRUE;
      sLastTokenStart = sBuffer + 1;
    }
    else if (bInToken)
    {
      if (isspace((unsigned char)*sBuffer))
        bInToken = FALSE;
    }
    else
    {
      if (!isspace((unsigned char)*sBuffer))
      {
        sLastTokenStart = sBuffer;
        bInToken = TRUE;
      }
    }
    ++sBuffer;
  }
  return sLastTokenStart;
}

/***********************************************************
* const char *Parse_SkipToken()                            *
*                         *  *  *  *                       *
*    Arguments:                                            *
*                                                          *
*    - const char *sText;                                  *
*                         *  *  *  *                       *
* Returns a pointer to the start of the second token in    *
* sText, or NULL.                                          *
***********************************************************/
const char *Parse_SkipToken(const char *sText)
{
  GetToken(sText, NULL, &sText, FALSE);
  return sText;
}


/***********************************************************
* unsigned Parse_ReplaceArgs()                             *
*                         *  *  *  *                       *
*    Arguments:                                            *
*                                                          *
*    - const char *Command;                                *
*      The Command template.                               *
*    - const char *Args;                                   *
*      The arguments to the command.                       *
*    - char *Buffer;                                       *
*      Where to put the result.                            *
*    - unsigned cchCapacity;                               *
*      Size of Buffer in chars.                            *
*                         *  *  *  *                       *
* Copies Command to the Buffer, replacing all occurances   *
* of the string "<args>" with an exact copy of Args.       *
***********************************************************/
unsigned Parse_ReplaceArgs(const char *Command, const char *Args, char *Buffer, unsigned cchCapacity)
{
  unsigned cchUsed = 0;
  const char *p;

  while (*Command)
  {
    /* look for replacement string */
    if (*Command == '<' && CSTR_EQUAL ==
      CompareString(LOCALE_USER_DEFAULT, NORM_IGNORECASE,
        Command, CT_STRLEN("<args>"), "<args>", CT_STRLEN("<args>")))
    {
      /* replace */ 
      Command += CT_STRLEN("<args>");

      p = Args;
      while (*p)
      {
        *Buffer++ = *p++;
        if (++cchUsed == cchCapacity)
          break;
      }
      if (cchUsed > cchCapacity)
        break;
    }
    else
    {
      if (IsDBCSLeadByte(*Command))
      {
        /* copy an extra char, ignoring '<' as a trail byte */ 
        *Buffer++ = *Command++;
        if (++cchUsed == cchCapacity)
          break;
      }
      *Buffer++ = *Command++;
      if (++cchUsed == cchCapacity)
        break;
    }
  }

  *Buffer = 0;
  if (cchUsed > cchCapacity)
    return 0;
  return cchUsed;
}
